# -*- coding: utf-8 -*-
#   StartUp Manager - Tool for editing settings for Grub and Usplash
#   Copyright (C) 2006-2007  Jimmy Rönnholm
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import os
import re
from subprocess import call

class Grub:

    """Represent the Grub configuration."""

    def __init__(self):
        """Set the variables"""
        self.config_file = '/etc/default/grub'
        self.header_config_file = '/etc/grub.d/00_header'
        self.readonly_config_file = '/boot/grub/grub.cfg'
        self.update_grub_command = '/usr/sbin/update-grub'
        self.grub_install_command = '/usr/sbin/grub-install'
        self.is_changed = False

        # Check that grub commands exist
        for path in [self.update_grub_command, self.grub_install_command,
                     self.config_file, self.header_config_file]:
            if not os.path.isfile(path):
                raise SystemExit('Fatal error. Can not find %s' % path)
#        call(self.update_grub_command) # Make sure config files are up to date

    def __read_lines_from_file(self, filename):
        """Read file filename, return list of lines in the file."""
        try:
            input_file = open(filename, 'r')
            lines = input_file.readlines()
            input_file.close()
            return lines
        except IOError:
            raise SystemExit('File %s does not exist.' % filename)

    def __write_lines_to_file(self, filename, lines):
        """Write the list lines to file filename."""
        output_file = open(filename, 'w')
        for out_line in lines:
            output_file.write(out_line)
        output_file.close()

    def __get_line_number(self, keyword):
        """Return number of the line where the value for keyword can be found.

        Return 1 if there is no match.

        """
        filename = self.__keyword_to_filename(keyword)
        identifier = self.__keyword_to_identifier(keyword)
        lines = self.__read_lines_from_file(filename)
        length = len(identifier)
        tracker = 0
        for line in lines:
            line = line.strip()[:length].strip()
            if line == identifier:
                return tracker
            tracker += 1
        return -1

    def __keyword_to_identifier(self, keyword):
        """Return an identifier to find the value for keyword.

        Return 1 if there is no match.
        Keyword can be:
        timeout
        entries
        default_boot
        vga_code
        show_text
        show_splash
        gfxmode

        """
        options = {
        'timeout'       : 'GRUB_TIMEOUT=',
        'entries'       : 'menuentry',
        'default_boot'  : 'GRUB_DEFAULT=',
        'vga_code'      : 'GRUB_CMDLINE_LINUX=',
        'show_text'     : 'GRUB_CMDLINE_LINUX=',
        'show_splash'   : 'GRUB_CMDLINE_LINUX=',
        'gfxmode'       : 'set gfxmode='
        }
        if options.has_key(keyword):
            return options[keyword]
        return 1

    def __keyword_to_filename(self, keyword):
        """Return the path to the file where value for keyword can be found.

        Return 1 if there is no match.
        Keyword can be:
        timeout
        entries
        default_boot
        vga_code
        show_text
        show_splash
        gfxmode

        """
        options = {
        'timeout'       : self.config_file,
        'entries'       : self.readonly_config_file,
        'default_boot'  : self.config_file,
        'vga_code'      : self.config_file,
        'show_text'     : self.config_file,
        'show_splash'   : self.config_file,
        'gfxmode'       : self.header_config_file
        }
        if options.has_key(keyword):
            return options[keyword]
        return 1

    def __get_boot_entries(self):
        """Return a list of the titles for the boot entries."""
        keyword = 'entries'
        identifier = self.__keyword_to_identifier(keyword)
        filename = self.__keyword_to_filename(keyword)
        lines = self.__read_lines_from_file(filename)
        entries = []
        for line in lines:
            if line[:len(identifier)] == identifier:
                first = line.find('"') + 1
                last = line.rfind('"')
                entries.append(line[first:last])
        return entries

    def __get_linux_options(self):
        """Return a list of the options passed to the Linux kernel."""
        keyword = 'vga_code'
        filename = self.__keyword_to_filename(keyword)
        lines = self.__read_lines_from_file(filename)
        line_number = self.__get_line_number(keyword)
        line = lines[line_number]
        first = line.find('"') + 1
        last = line.rfind('"')
        return line[first:last].split(' ')

    def __read_config(self, keyword):
        """Read grub configuration files, return value for keyword.

        Return -1 if value cannot be found.
        Keyword can be:
        timeout
        entries
        default_boot
        vga_code
        show_text
        show_splash
        gfxmode

        """
        identifier = self.__keyword_to_identifier(keyword)
        filename = self.__keyword_to_filename(keyword)
        line_number = self.__get_line_number(keyword)
        if line_number == -1:
            return line_number
        lines = self.__read_lines_from_file(filename)
        line = lines[line_number]
        length = len(identifier)
        line = line[length:]
        if keyword == 'entries':
            return self.__get_boot_entries()

        if identifier == 'GRUB_CMDLINE_LINUX=':
            options = self.__get_linux_options()
            if keyword == 'vga_code':
                for option in options:
                    if option[:4] == 'vga=':
                        value = option[4:]
                        if len(value) == 5:
                            # Convert from hexadecimal format
                            return int(value, 16)
                        return int(value)

            if keyword == 'show_text':
                if not 'quiet' in options:
                    return True
                return False
        
            if keyword == 'show_splash':
                if 'splash' in options:
                    return True
                return False

        if keyword == 'timeout' or keyword == 'default_boot':
            number_filter = re.compile('[0-9]+')
            match = number_filter.search(line)
            if match:
                return int(match.group())

        if keyword == 'gfxmode':
            return line.strip()

        return -1

    def __write_config(self, keyword, value):
        """Change the configuration files based on the passed value to keyword.

        Return 1 if the operation failed.
        Keyword and value can be:
        timeout         integer
        default_boot    integer
        vga_code        integer, grub vga code
        show_text       boolean
        show_splash     boolean
        gfxmode         string resolution ex. '1280x1024'

        """
        identifier = self.__keyword_to_identifier(keyword)
        file_name = self.__keyword_to_filename(keyword)
        line_number = self.__get_line_number(keyword)
        lines = self.__read_lines_from_file(file_name)

        if line_number == -1:
            if keyword == 'gfxmode':
                # We do not know where to insert the value
                return 1
            lines.append("\n")
            line_number = len(lines) - 1

        if identifier == 'GRUB_CMDLINE_LINUX=':
            options = self.__get_linux_options()

            if keyword == 'vga_code':
                for i, option in enumerate(options):
                    if option[:4] == 'vga=':
                        del options[i]
                options.append('vga=' + str(value))

            if keyword == 'show_text':
                if not value:
                    options.append('quiet')
                elif 'quiet' in options:
                    options.remove('quiet')

            if keyword == 'show_splash':
                if value:
                    options.append('splash')
                elif 'splash' in options:
                    options.remove('splash')

            options = set(options)
            new_line = 'GRUB_CMDLINE_LINUX="'
            for option in options:
                new_line += option + ' '
            
            new_line = new_line[:-1] + '"\n' # Remove the last space

        else:
            new_line = identifier + str(value) + '\n'

        lines[line_number] = new_line
        self.__write_lines_to_file(file_name, lines)
        self.is_changed = True

    def get_timeout(self):
        """Return the timeout in seconds used for Grub menu."""
        timeout = self.__read_config('timeout')
        if timeout == -1:
            timeout = 5
        return timeout

    def set_timeout(self, timeout):
        """Set the timeout in seconds used in the Grub menu.

        timeout = number of seconds

        """
        self.__write_config('timeout', timeout)

    def get_titles(self):
        """Return names of the boot options, as a list."""
        return self.__read_config('entries')

    def get_default_boot(self):
        """Return the number of the default booted option."""
        default_boot = self.__read_config('default_boot')
        if default_boot == -1:
            default_boot = 0
        return default_boot

    def set_default_boot(self, default):
        """Set the number of the default booted option."""
        self.__write_config('default_boot', default)

    def get_vga_code(self):
        """Return the Grub vga code used, as integer.

        If no code is specified in the config, 769 is returned.
        Grub vga codes:

        Colours   640x400 640x480 800x600 1024x768 1152x864 1280x1024 1600x1200
        --------+--------------------------------------------------------------
         4 bits |    ?       ?      770       ?        ?        ?         ?
         8 bits |   768     769     771      773      353      775       796 
        15 bits |    ?      784     787      790      354      793       797 
        16 bits |    ?      758     788      791      355      794       798 
        24 bits |    ?      786     789      792       ?       795       799 
        32 bits |    ?       ?       ?        ?       356       ?

        """
        vga_code = self.__read_config('vga_code')
        if vga_code == -1:
            vga_code = 769
        return vga_code

    def set_vga_code(self, vga):
        """Set the resolution based on Grub vga code.

        vga = grub vga code
        For vga codes, see get_vga_code

        """
        self.__write_config('vga_code', vga)

    def get_boot_text_visible(self):
        """Return whether a verbose boot is enabled."""
        show_text = self.__read_config('show_text')
        if show_text == -1:
            show_text = True
        return show_text

    def set_boot_text_visible(self, active):
        """Set whether a verbose boot is enabled."""
        self.__write_config('show_text', active)

    def get_splash_active(self):
        """Return whether a boot splash is shown."""
        show_splash = self.__read_config('show_splash')
        if show_splash == -1:
            show_splash = False
        return show_splash

    def set_splash_active(self, active):
        """Set whether a boot splash is shown."""
        self.__write_config('show_splash', active)

    def get_gfxmode(self):
        """Return the gfxmode resolution used in Grub menu.

        Return -1 if no gfxmode can be read.

        """
        return self.__read_config('gfxmode')

    def set_gfxmode(self, resolution):
        """Set the gfxmode resolution for Grub menu. 

        resolution ex. '1280x1024'
        Return -1 if the value can not be set.

        """
        self.__write_config('gfxmode', resolution)

    def format_floppy(self):
        """Format a floppy.

        Return 0 if it was successful.
        Return 1 if a floppy was not found.
        Return 3 if there was an OSError.
        Return a code < 0 or > 3 for other errors.

        """
        try:
            retcode = call('/sbin/mke2fs /dev/fd0')
            if retcode < 0:
                print 'Child was terminated by signal', -retcode
            return retcode
        except OSError, e:
            print 'Execution failed:', e
            return 3

    def make_floppy(self):
        """Write a Grub boot floppy.

        Return 0 if it was successful.
        Return 1 if there was an OSError.
        Return a code < 0 or > 1 for other errors.

        """
        dir_preexisting = True

        if not os.path.isdir('/mnt/floppysum'):
            os.mkdir('/mnt/floppysum')
            dir_preexisting = False
        try:
            retcode = call('mount -t ext2 /dev/fd0 /mnt/floppysum')
            if not retcode == 0:
                print 'Child was terminated by signal', -retcode
                return retcode

            retcode = call(self.grub_install_command + 
                           ' --root-directory=/mnt/floppysum fd0')
            if not retcode == 0:
                print 'Child was terminated by signal', -retcode
                return retcode

            retcode = call('cp ' + self.config_file + '  /mnt/floppysum/' + 
                           self.config_file)
            if not retcode == 0:
                print 'Child was terminated by signal', -retcode
                return retcode

            retcode = call('umount /mnt/floppysum')
            if not retcode == 0:
                print 'Child was terminated by signal', -retcode
                return retcode

        except OSError, e:
            print 'Execution failed:', e
            return 1

        if not dir_preexisting:
            os.rmdir('/mnt/floppysum')

        return 0

    def do_shutdown_tasks(self):
        """Do any post-config tasks necessary."""
        if self.is_changed:
            call(self.update_grub_command)

